@@ -3,36 +3,31 @@
  Maintainer: Thomas.DuBuisson@gmail.com
  Stability: beta
  Portability: portable 
- Authors: Thomas DuBuisson, Francisco Blas Izquierdo Riera (klondike)
+ Authors: Thomas DuBuisson
 
 
  Generic mode implementations useable by any correct BlockCipher
-instance Be aware there are no tests for CFB mode yet.  See
-'Test.Crypto'. 
+ instance Be aware there are no tests for CFB mode yet.  See
+ 'Test.Crypto'.
 -}
 module Crypto.Modes (
-	-- * Initialization Vector Type, Modifiers (for all ciphers, all modes that use IVs)
-	  IV
-	, getIV, getIVIO, zeroIV
-	, incIV, dblIV
-	-- * Blockcipher modes. Names with a prime (') means strict, without a prime means lazy bytestrings.
-	, ecb, unEcb
-	, cbc, unCbc
-	, cfb, unCfb
-	, ofb, unOfb
-	, ecb', unEcb'
-	, cbc', unCbc'
-	, cfb', unCfb'
-	, ofb', unOfb'
-	, ctr, unCtr, ctr', unCtr'
-	, siv, unSiv, siv', unSiv'
-	-- * Authentication modes
-	, cbcMac', cbcMac, cMac, cMac'
-	-- * Combined modes (nothing here yet)
-	-- , gmc
-	-- , xts
-	-- , ccm
-	) where
+        -- * Initialization Vector Type, Modifiers (for all ciphers, all modes that use IVs)
+          getIV, getIVIO, zeroIV
+        , dblIV
+        -- * Blockcipher modes for lazy bytestrings. Versions for strict bytestrings are in 'Crypto.Classes'.
+        , Crypto.Modes.ecb, Crypto.Modes.unEcb
+        , Crypto.Modes.cbc, Crypto.Modes.unCbc
+        , Crypto.Modes.cfb, Crypto.Modes.unCfb
+        , Crypto.Modes.ofb, Crypto.Modes.unOfb
+        , Crypto.Modes.ctr, Crypto.Modes.unCtr
+        , siv, unSiv, siv', unSiv'
+        -- * Authentication modes
+        , cbcMac', cbcMac, cMac, cMac' 
+        -- * Combined modes (nothing here yet)
+        -- , gmc
+        -- , xts
+        -- , ccm
+        ) where
 
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as L
@@ -41,10 +36,11 @@ import qualified Data.Serialize.Put as SP
 import qualified Data.Serialize.Get as SG
 import Data.Bits (xor, shift, (.&.), (.|.), testBit, setBit, clearBit, Bits, complementBit)
 import Data.Tagged
-import Crypto.Classes
+import Crypto.Classes (BlockCipher(..), for, blockSizeBytes, incIV)
 import Crypto.Random
 import Crypto.Util
 import Crypto.CPoly
+import Crypto.Types
 import System.Entropy (getEntropy)
 import Control.Monad (liftM, forM_)
 import Data.List (genericDrop)
@@ -55,22 +51,12 @@ import Data.List (genericDrop,genericReplicate,genericLength)
 import Data.Proxy
 #endif
 
--- |Initilization Vectors for BlockCipher implementations (IV k) are
--- used for various modes and guarrenteed to be blockSize bits long.
--- The common ways to obtain an IV are to generate one ('getIV' or
--- 'getIVIO') or to use one provided with the ciphertext (using the
--- 'Serialize' instance of IV).
---
--- 'zeroIV' also exists and is of particular use for starting 'ctr'
--- mode with a fresh key.
-data IV k = IV { initializationVector :: {-# UNPACK #-} !B.ByteString } deriving (Eq, Ord, Show)
-
 -- gather a specified number of bytes from the list of bytestrings
 collect :: Int -> [B.ByteString] -> [B.ByteString]
 collect 0 _ = []
 collect _ [] = []
 collect i (b:bs)
-	| len < i  = b : collect (i - len) bs
+        | len < i  = b : collect (i - len) bs
         | len >= i = [B.take i b]
   where
   len = B.length b
@@ -101,42 +87,33 @@ chunkFor' k = go
 -- than either of the two inputs.
 zwp :: L.ByteString -> L.ByteString -> L.ByteString
 zwp  a b = 
-	let as = L.toChunks a
-	    bs = L.toChunks b
-	in L.fromChunks (go as bs)
+        let as = L.toChunks a
+            bs = L.toChunks b
+        in L.fromChunks (go as bs)
   where
   go [] _ = []
   go _ [] = []
   go (a:as) (b:bs) =
-	let l = min (B.length a) (B.length b)
-	    (a',ar) = B.splitAt l a
-	    (b',br) = B.splitAt l b
-	    as' = if B.length ar == 0 then as else ar : as
-	    bs' = if B.length br == 0 then bs else br : bs
-	in (zwp' a' b') : go as' bs'
+        let l = min (B.length a) (B.length b)
+            (a',ar) = B.splitAt l a
+            (b',br) = B.splitAt l b
+            as' = if B.length ar == 0 then as else ar : as
+            bs' = if B.length br == 0 then bs else br : bs
+        in (zwp' a' b') : go as' bs'
 {-# INLINEABLE zwp #-}
 
--- |zipWith xor + Pack
---
--- As a result of rewrite rules, this should automatically be
--- optimized (at compile time) to use the bytestring libraries
--- 'zipWith'' function.
-zwp' :: B.ByteString -> B.ByteString -> B.ByteString
-zwp' a = B.pack . B.zipWith xor a
-{-# INLINEABLE zwp' #-}
-
 -- |Cipher block chaining encryption mode on strict bytestrings
 cbc' :: BlockCipher k => k -> IV k -> B.ByteString -> (B.ByteString, IV k)
 cbc' k (IV v) plaintext =
-	let blks = chunkFor' k plaintext
-	    (cts, iv) = go blks v
-	in (B.concat cts, IV iv)
+        let blks = chunkFor' k plaintext
+            (cts, iv) = go blks v
+        in (B.concat cts, IV iv)
   where
   go [] iv = ([], iv)
   go (b:bs) iv =
-	let c = encryptBlock k (zwp' iv b)
-	    (cs, ivFinal) = go bs c
-	in (c:cs, ivFinal)
+        let c = encryptBlock k (zwp' iv b)
+            (cs, ivFinal) = go bs c
+        in (c:cs, ivFinal)
 {-# INLINEABLE cbc' #-}
 
 -- |Cipher block chaining message authentication
@@ -146,54 +123,54 @@ cbcMac' k pt = encode $ snd $ cbc' k zeroIV pt
 
 -- |Cipher block chaining message authentication
 cbcMac :: BlockCipher k => k -> L.ByteString -> L.ByteString
-cbcMac k pt = L.fromChunks [encode $ snd $ cbc k zeroIV pt]
+cbcMac k pt = L.fromChunks [encode $ snd $ Crypto.Modes.cbc k zeroIV pt]
 {-# INLINEABLE cbcMac #-}
 
 -- |Cipher block chaining decryption for strict bytestrings
 unCbc' :: BlockCipher k => k -> IV k -> B.ByteString -> (B.ByteString, IV k)
 unCbc' k (IV v) ciphertext =
-	let blks = chunkFor' k ciphertext
-	    (pts, iv) = go blks v
-	in (B.concat pts, IV iv)
+        let blks = chunkFor' k ciphertext
+            (pts, iv) = go blks v
+        in (B.concat pts, IV iv)
   where
   go [] iv = ([], iv)
   go (c:cs) iv =
-	let p = zwp' (decryptBlock k c) iv
-	    (ps, ivFinal) = go cs c
-	in (p:ps, ivFinal)
+        let p = zwp' (decryptBlock k c) iv
+            (ps, ivFinal) = go cs c
+        in (p:ps, ivFinal)
 {-# INLINEABLE unCbc' #-}
 
 -- |Cipher block chaining encryption for lazy bytestrings
 cbc :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
 cbc k (IV v) plaintext =
-	let blks = chunkFor k plaintext
-	    (cts, iv) = go blks v
-	in (L.fromChunks cts, IV iv)
+        let blks = chunkFor k plaintext
+            (cts, iv) = go blks v
+        in (L.fromChunks cts, IV iv)
   where
   go [] iv = ([], iv)
   go (b:bs) iv =
-	let c = encryptBlock k (zwp' iv b)
-	    (cs, ivFinal) = go bs c
-	in (c:cs, ivFinal)
+        let c = encryptBlock k (zwp' iv b)
+            (cs, ivFinal) = go bs c
+        in (c:cs, ivFinal)
 {-# INLINEABLE cbc #-}
 
 -- |Cipher block chaining decryption for lazy bytestrings
 unCbc :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
 unCbc k (IV v) ciphertext =
-	let blks = chunkFor k ciphertext
-	    (pts, iv) = go blks v
-	in (L.fromChunks pts, IV iv)
+        let blks = chunkFor k ciphertext
+            (pts, iv) = go blks v
+        in (L.fromChunks pts, IV iv)
   where
   go [] iv = ([], iv)
   go (c:cs) iv =
-	let p = zwp' (decryptBlock k c) iv
-	    (ps, ivFinal) = go cs c
-	in (p:ps, ivFinal)
+        let p = zwp' (decryptBlock k c) iv
+            (ps, ivFinal) = go cs c
+        in (p:ps, ivFinal)
 {-# INLINEABLE unCbc #-}
 
 -- |Counter mode for lazy bytestrings
 ctr :: BlockCipher k => (IV k -> IV k) -> k -> IV k -> L.ByteString -> (L.ByteString, IV k)
-ctr = unCtr
+ctr = Crypto.Modes.unCtr
 
 -- |Counter  mode for lazy bytestrings
 unCtr :: BlockCipher k => (IV k -> IV k) -> k -> IV k -> L.ByteString -> (L.ByteString, IV k)
@@ -203,18 +180,6 @@ unCtr f k (IV iv) msg =
            newIV = head $ genericDrop ((ivLen - 1 + L.length msg) `div` ivLen) ivStr
        in (zwp (L.fromChunks $ map (encryptBlock k) $ map initializationVector ivStr) msg, newIV)
 
--- |Counter mode for strict bytestrings
-ctr' :: BlockCipher k => (IV k -> IV k) -> k -> IV k -> B.ByteString -> (B.ByteString, IV k)
-ctr' = unCtr'
-
--- |Counter mode for strict bytestrings
-unCtr' :: BlockCipher k => (IV k -> IV k) -> k -> IV k -> B.ByteString -> (B.ByteString, IV k)
-unCtr' f k (IV iv) msg =
-       let ivStr = iterate f $ IV iv
-           ivLen = fromIntegral $ B.length iv
-           newIV = head $ genericDrop ((ivLen - 1 + B.length msg) `div` ivLen) ivStr
-       in (zwp' (B.concat $ collect (B.length msg) (map (encryptBlock k . initializationVector) ivStr)) msg, newIV)
-
 -- |Generate cmac subkeys.  The usage of seq tries to force evaluation
 -- of both keys avoiding posible timing attacks
 cMacSubk :: BlockCipher k => k -> (IV k, IV k)
@@ -277,10 +242,13 @@ xorend bsize (n, x:xs) | n <= bsize = Just (x,((n-1),xs))
 cMacStar :: BlockCipher k => k -> [L.ByteString] -> L.ByteString
 cMacStar k l = go (lcmac (L.replicate bSize 0)) l
   where
-       bSize = fromIntegral $ blockSizeBytes `for` k
-       bSizeb = fromIntegral $ blockSize `for` k
-       lcmac = cMacWithSubK k (cMacSubk k)
-       go s [] = s
+        bSize = fromIntegral $ blockSizeBytes `for` k
+        bSizeb = fromIntegral $ blockSize `for` k
+        lcmac = cMacWithSubK k (cMacSubk k)
+        go s [] = s
+        go s [x] | (L.length x) >= bSize = lcmac $ zwp x $ L.unfoldr (xorend $ fromIntegral bSize) (fromIntegral $ L.length x,L.unpack s)
+                 | otherwise = lcmac $ zwp (dblL s) (L.unfoldr cMacPad (L.unpack x,True,fromIntegral bSize))
+        go s (x:xs) = go (zwp (dblL s) (lcmac x)) xs
 
 -- |Obtain the CMAC* on strict bytestrings
 cMacStar' :: BlockCipher k => k -> [B.ByteString] -> B.ByteString
@@ -292,9 +260,7 @@ cMacStar' k s = go (lcmac (B.replicate bSize 0)) s
        go s [] = s
        go s [x] | (B.length x) >= bSize = lcmac $ zwp' x $ fst $ B.unfoldrN (B.length x) (xorend bSize) (fromIntegral $ B.length x,B.unpack s)
                 | otherwise = lcmac $ zwp' (dblB s) (fst $ B.unfoldrN bSize cMacPad (B.unpack x,True,bSize))
-       go s (x:xs) = go (zwp' (dblB s) (lcmac x)) xs
-
-
+       go s (x:xs) = go (zwp' (dblB s) (lcmac x)) xs 
 
 -- |Create the mask for SIV based ciphers
 sivMask :: B.ByteString -> B.ByteString
@@ -311,7 +277,7 @@ sivMask b = snd $ B.mapAccumR (go) 0 b
 -- return nothing when certain constraints aren't met.
 siv :: BlockCipher k => k -> k -> [L.ByteString] -> L.ByteString -> Maybe L.ByteString
 siv k1 k2 xs m | length xs > bSizeb - 1 = Nothing
-               | otherwise = Just $ L.append iv $ fst $ ctr incIV k2 (IV $ sivMask $ B.concat $ L.toChunks iv) m
+               | otherwise = Just $ L.append iv $ fst $ Crypto.Modes.ctr incIV k2 (IV $ sivMask $ B.concat $ L.toChunks iv) m
   where
        bSize = fromIntegral $ blockSizeBytes `for` k1
        bSizeb = fromIntegral $ blockSize `for` k1
@@ -331,7 +297,7 @@ unSiv k1 k2 xs c | length xs > bSizeb - 1 = Nothing
        bSize = fromIntegral $ blockSizeBytes `for` k1
        bSizeb = fromIntegral $ blockSize `for` k1
        (iv,m) = L.splitAt (fromIntegral bSize) c
-       dm = fst $ unCtr incIV k2 (IV $ sivMask $ B.concat $ L.toChunks iv) m
+       dm = fst $ Crypto.Modes.unCtr incIV k2 (IV $ sivMask $ B.concat $ L.toChunks iv) m
 
 -- |SIV (Synthetic IV) mode for strict bytestrings.  First argument is
 -- the optional list of bytestrings to be authenticated but not
@@ -339,14 +305,12 @@ unSiv k1 k2 xs c | length xs > bSizeb - 1 = Nothing
 -- return nothing when certain constraints aren't met.
 siv' :: BlockCipher k => k -> k -> [B.ByteString] -> B.ByteString -> Maybe B.ByteString
 siv' k1 k2 xs m | length xs > bSizeb - 1 = Nothing
-                | otherwise = Just $ B.append iv $ fst $ ctr' incIV k2 (IV $ sivMask iv) m
+                | otherwise = Just $ B.append iv $ fst $ Crypto.Classes.ctr k2 (IV $ sivMask iv) m
   where
        bSize = fromIntegral $ blockSizeBytes `for` k1
        bSizeb = fromIntegral $ blockSize `for` k1
        iv = cMacStar' k1 $ xs ++ [m]
 
-
-
 -- |SIV (Synthetic IV) for strict bytestrings First argument is the
 -- optional list of bytestrings to be authenticated but not encrypted
 -- As required by the specification this algorithm may return nothing
@@ -360,16 +324,7 @@ unSiv' k1 k2 xs c | length xs > bSizeb - 1 = Nothing
        bSize = fromIntegral $ blockSizeBytes `for` k1
        bSizeb = fromIntegral $ blockSize `for` k1
        (iv,m) = B.splitAt bSize c
-       dm = fst $ unCtr' incIV k2 (IV $ sivMask iv) m
-
--- |Increase an `IV` by one.  This is way faster than decoding,
--- increasing, encoding
-incIV :: BlockCipher k => IV k -> IV k
-incIV (IV b) = IV $ snd $ B.mapAccumR (incw) True b
-  where
-       incw :: Bool -> Word8 -> (Bool, Word8)
-       incw True w = (w == maxBound, w + 1)
-       incw False w = (False, w)
+       dm = fst $ Crypto.Classes.unCtr k2 (IV $ sivMask iv) m
 
 -- |Accumulator based double operation
 dblw :: Bool -> (Int,[Int],Bool) -> Word8 -> ((Int,[Int],Bool), Word8)
@@ -387,7 +342,6 @@ dblw hb (i,xs,b) w = dblw' hb
        dblw' False = i'`seq`xs'`seq`w''`seq`((i,xs,b'),w')
        dblw' True  = ((i',xs',b'),w'')
 
-
 -- |Perform doubling as defined by the CMAC and SIV papers
 dblIV :: BlockCipher k => IV k -> IV k
 dblIV (IV b) = IV $ dblB b
@@ -401,7 +355,7 @@ dblB b | B.null b = b
 dblL :: L.ByteString -> L.ByteString
 dblL b | L.null b = b
        | otherwise = snd $ L.mapAccumR (dblw (testBit (L.head b) 7)) (0,cpoly2revlist (L.length b * 8),False) b
-
+ 
 -- |Cast a bigEndian ByteString into an Integer
 decodeB :: B.ByteString -> Integer
 decodeB = B.foldl' (\acc w -> (shift acc 8) + toInteger(w)) 0
@@ -449,102 +403,102 @@ zeroIV = iv
 -- |Cook book mode - not really a mode at all.  If you don't know what you're doing, don't use this mode^H^H^H^H library.
 ecb :: BlockCipher k => k -> L.ByteString -> L.ByteString
 ecb k msg =
-	let chunks = chunkFor k msg
-	in L.fromChunks $ map (encryptBlock k) chunks
+        let chunks = chunkFor k msg
+        in L.fromChunks $ map (encryptBlock k) chunks
 {-# INLINEABLE ecb #-}
 
 -- |ECB decrypt, complementary to `ecb`.
 unEcb :: BlockCipher k => k -> L.ByteString -> L.ByteString
 unEcb k msg =
-	let chunks = chunkFor k msg
-	in L.fromChunks $ map (decryptBlock k) chunks
+        let chunks = chunkFor k msg
+        in L.fromChunks $ map (decryptBlock k) chunks
 {-# INLINEABLE unEcb #-}
 
 -- | Like `ecb` but for strict bytestrings
 ecb' :: BlockCipher k => k -> B.ByteString -> B.ByteString
 ecb' k msg =
-	let chunks = chunkFor' k msg
-	in B.concat $ map (encryptBlock k) chunks
+        let chunks = chunkFor' k msg
+        in B.concat $ map (encryptBlock k) chunks
 {-# INLINEABLE ecb' #-}
 
 -- |Decryption complement to `ecb'`
 unEcb' :: BlockCipher k => k -> B.ByteString -> B.ByteString
 unEcb' k ct =
-	let chunks = chunkFor' k ct
-	in B.concat $ map (decryptBlock k) chunks
+        let chunks = chunkFor' k ct
+        in B.concat $ map (decryptBlock k) chunks
 {-# INLINEABLE unEcb' #-}
 
 -- |Ciphertext feed-back encryption mode for lazy bytestrings (with s
 -- == blockSize)
 cfb :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
 cfb k (IV v) msg =
-	let blks = chunkFor k msg
-	    (cs,ivF) = go v blks
-	in (L.fromChunks cs, IV ivF)
+        let blks = chunkFor k msg
+            (cs,ivF) = go v blks
+        in (L.fromChunks cs, IV ivF)
   where
   go iv [] = ([],iv)
   go iv (b:bs) =
-	let c = zwp' (encryptBlock k iv) b
-	    (cs,ivFinal) = go c bs
-	in (c:cs, ivFinal)
+        let c = zwp' (encryptBlock k iv) b
+            (cs,ivFinal) = go c bs
+        in (c:cs, ivFinal)
 {-# INLINEABLE cfb #-}
 
 -- |Ciphertext feed-back decryption mode for lazy bytestrings (with s
 -- == blockSize)
 unCfb :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
 unCfb k (IV v) msg = 
-	let blks = chunkFor k msg
-	    (ps, ivF) = go v blks
-	in (L.fromChunks ps, IV ivF)
+        let blks = chunkFor k msg
+            (ps, ivF) = go v blks
+        in (L.fromChunks ps, IV ivF)
   where
   go iv [] = ([], iv)
   go iv (b:bs) =
-	let p = zwp' (encryptBlock k iv) b
-	    (ps, ivF) = go b bs
-	in (p:ps, ivF)
+        let p = zwp' (encryptBlock k iv) b
+            (ps, ivF) = go b bs
+        in (p:ps, ivF)
 {-# INLINEABLE unCfb #-}
 
 -- |Ciphertext feed-back encryption mode for strict bytestrings (with
 -- s == blockSize)
 cfb' :: BlockCipher k => k -> IV k -> B.ByteString -> (B.ByteString, IV k)
 cfb' k (IV v) msg =
-	let blks = chunkFor' k msg
-	    (cs,ivF) = go v blks
-	in (B.concat cs, IV ivF)
+        let blks = chunkFor' k msg
+            (cs,ivF) = go v blks
+        in (B.concat cs, IV ivF)
   where
   go iv [] = ([],iv)
   go iv (b:bs) =
-	let c = zwp' (encryptBlock k iv) b
-	    (cs,ivFinal) = go c bs
-	in (c:cs, ivFinal)
+        let c = zwp' (encryptBlock k iv) b
+            (cs,ivFinal) = go c bs
+        in (c:cs, ivFinal)
 {-# INLINEABLE cfb' #-}
 
 -- |Ciphertext feed-back decryption mode for strict bytestrings (with s == blockSize)
 unCfb' :: BlockCipher k => k -> IV k -> B.ByteString -> (B.ByteString, IV k)
 unCfb' k (IV v) msg =
-	let blks = chunkFor' k msg
-	    (ps, ivF) = go v blks
-	in (B.concat ps, IV ivF)
+        let blks = chunkFor' k msg
+            (ps, ivF) = go v blks
+        in (B.concat ps, IV ivF)
   where
   go iv [] = ([], iv)
   go iv (b:bs) =
-	let p = zwp' (encryptBlock k iv) b
-	    (ps, ivF) = go b bs
-	in (p:ps, ivF)
+        let p = zwp' (encryptBlock k iv) b
+            (ps, ivF) = go b bs
+        in (p:ps, ivF)
 {-# INLINEABLE unCfb' #-}
 
 -- |Output feedback mode for lazy bytestrings
 ofb :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
-ofb = unOfb
+ofb = Crypto.Modes.unOfb
 {-# INLINEABLE ofb #-}
 
 -- |Output feedback mode for lazy bytestrings
 unOfb :: BlockCipher k => k -> IV k -> L.ByteString -> (L.ByteString, IV k)
 unOfb k (IV iv) msg =
-	let ivStr = drop 1 (iterate (encryptBlock k) iv)
-	    ivLen = fromIntegral (B.length iv)
-	    newIV = IV . B.concat . L.toChunks . L.take ivLen . L.drop (L.length msg) . L.fromChunks $ ivStr
-	in (zwp (L.fromChunks ivStr) msg, newIV)
+        let ivStr = drop 1 (iterate (encryptBlock k) iv)
+            ivLen = fromIntegral (B.length iv)
+            newIV = IV . B.concat . L.toChunks . L.take ivLen . L.drop (L.length msg) . L.fromChunks $ ivStr
+        in (zwp (L.fromChunks ivStr) msg, newIV)
 {-# INLINEABLE unOfb #-}
 
 -- |Output feedback mode for strict bytestrings
@@ -555,35 +509,35 @@ ofb' = unOfb'
 -- |Output feedback mode for strict bytestrings
 unOfb' :: BlockCipher k => k -> IV k -> B.ByteString -> (B.ByteString, IV k)
 unOfb' k (IV iv) msg =
-	let ivStr = collect (B.length msg + ivLen) (drop 1 (iterate (encryptBlock k) iv))
-	    ivLen = B.length iv
-	    mLen = fromIntegral (B.length msg)
-	    newIV = IV . B.concat . L.toChunks . L.take (fromIntegral ivLen) . L.drop mLen . L.fromChunks $ ivStr
-	in (zwp' (B.concat ivStr) msg, newIV)
+        let ivStr = collect (B.length msg + ivLen) (drop 1 (iterate (encryptBlock k) iv))
+            ivLen = B.length iv
+            mLen = fromIntegral (B.length msg)
+            newIV = IV . B.concat . L.toChunks . L.take (fromIntegral ivLen) . L.drop mLen . L.fromChunks $ ivStr
+        in (zwp' (B.concat ivStr) msg, newIV)
 {-# INLINEABLE unOfb' #-}
 
 -- |Obtain an `IV` using the provided CryptoRandomGenerator.
 getIV :: (BlockCipher k, CryptoRandomGen g) => g -> Either GenError (IV k, g)
 getIV g =
-	let bytes = ivBlockSizeBytes iv
-	    gen = genBytes bytes g
-	    fromRight (Right x) = x
-	    iv  = IV (fst  . fromRight $ gen)
-	in case gen of
-		Left err -> Left err
-		Right (bs,g')
-			| B.length bs == bytes	-> Right (iv, g')
-			| otherwise		-> Left (GenErrorOther "Generator failed to provide requested number of bytes")
+        let bytes = ivBlockSizeBytes iv
+            gen = genBytes bytes g
+            fromRight (Right x) = x
+            iv  = IV (fst  . fromRight $ gen)
+        in case gen of
+                Left err -> Left err
+                Right (bs,g')
+                        | B.length bs == bytes  -> Right (iv, g')
+                        | otherwise             -> Left (GenErrorOther "Generator failed to provide requested number of bytes")
 {-# INLINEABLE getIV #-}
 
 -- | Obtain an 'IV' using the system entropy (see 'System.Crypto.Random')
 getIVIO :: (BlockCipher k) => IO (IV k)
 getIVIO = do
-	let p = Proxy
-	    getTypedIV :: BlockCipher k => Proxy k -> IO (IV k)
-	    getTypedIV pr = liftM IV (getEntropy (proxy blockSize pr `div` 8))
-	iv <- getTypedIV p
-	return (iv `asProxyTypeOf` ivProxy p)
+        let p = Proxy
+            getTypedIV :: BlockCipher k => Proxy k -> IO (IV k)
+            getTypedIV pr = liftM IV (getEntropy (proxy blockSize pr `div` 8))
+        iv <- getTypedIV p
+        return (iv `asProxyTypeOf` ivProxy p)
 {-# INLINEABLE getIVIO #-}
 
 ivProxy :: Proxy k -> Proxy (IV k)
@@ -597,18 +551,18 @@ proxyOf = const Proxy
 
 ivBlockSizeBytes :: BlockCipher k => IV k -> Int
 ivBlockSizeBytes iv =
-	let p = deIVProxy (proxyOf iv)
-	in proxy blockSize p `div` 8
+        let p = deIVProxy (proxyOf iv)
+        in proxy blockSize p `div` 8
 {-# INLINEABLE ivBlockSizeBytes #-}
 
 instance (BlockCipher k) => Serialize (IV k) where
-	get = do
-		let p = Proxy
-		    doGet :: BlockCipher k => Proxy k -> Get (IV k)
-	            doGet pr = liftM IV (SG.getByteString (proxy blockSizeBytes pr))
-		iv <- doGet p
-		return (iv `asProxyTypeOf` ivProxy p)
-	put (IV iv) = SP.putByteString iv
+        get = do
+                let p = Proxy
+                    doGet :: BlockCipher k => Proxy k -> Get (IV k)
+                    doGet pr = liftM IV (SG.getByteString (proxy blockSizeBytes pr))
+                iv <- doGet p
+                return (iv `asProxyTypeOf` ivProxy p)
+        put (IV iv) = SP.putByteString iv
 
 -- TODO: GCM, GMAC
 -- Consider the AES-only modes of XTS, CCM
